/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#define LOG_CAM_FRAME_DROPS false

#include "ExpCamera.h"

#include "MixCast.h"
#include "MixCastSdkActor.h"
#include "Data/SharedTexIds.h"
#include "Transmission/Thrift/ServiceCommunicator.h"
#include "Transmission/Thrift/ThriftUtil.hpp"
#include "Utility/MathUtil.hpp"
#include "Utility/StringUtils.hpp"
#include "Utility/TimeUtils.hpp"
#include "ExpFrameSender.h"
#include "MixCastLog.h"

#include "Engine/Engine.h"
#include "GameFramework/WorldSettings.h"
#include "Camera/CameraComponent.h"
#include "Components/DrawFrustumComponent.h"
#include "Components/SceneCaptureComponent2D.h"
#include "Kismet/GameplayStatics.h"
#include "TimerManager.h"

#include <string>

AExpCamera::FFrameStartedEvent AExpCamera::FrameStarted;
AExpCamera::FFrameEndedEvent AExpCamera::FrameEnded;

TMap<FString, SharedQueueConsumer<ExpFrame>*> AExpCamera::cachedFrameQueues;

void AExpCamera::DisposeAllCachedQueues()
{
	for (auto& Elem : cachedFrameQueues)
		delete Elem.Value;
	cachedFrameQueues.Reset();
}

AExpCamera::AExpCamera(const FObjectInitializer& ObjectInitializer)
{
	UPluginProjectSettings* ProjectSettings = SdkData::GetProjectSettings();

	SpawnCollisionHandlingMethod = ESpawnActorCollisionHandlingMethod::AlwaysSpawn;

	CameraTransform = CreateDefaultSubobject<USceneComponent>(FName("Transform"));
	SetRootComponent(CameraTransform);

	FullCapture = CreateCaptureComponent(false);
	onPreFullRender = FullCapture->OnCaptureBegin().AddUObject(this, &AExpCamera::HandleSceneCaptureStarted);
	onPostFullRender = FullCapture->OnCaptureEnd().AddUObject(this, &AExpCamera::HandleSceneCaptureEnded);

	ForegroundCapture = CreateCaptureComponent(true);
	onPreForegroundRender = ForegroundCapture->OnCaptureBegin().AddUObject(this, &AExpCamera::HandleSceneCaptureStarted);
	onPostForegroundRender = ForegroundCapture->OnCaptureEnd().AddUObject(this, &AExpCamera::HandleSceneCaptureEnded);
}

void AExpCamera::BeginPlay()
{
	Super::BeginPlay();

	sdkModule = (FMixCast*)&IMixCast::Get();

	if (FullCapture->CaptureSource == ESceneCaptureSource::SCS_SceneColorHDRNoAlpha ||
		FullCapture->CaptureSource == ESceneCaptureSource::SCS_SceneColorSceneDepth)
	{
		FullCapture->CaptureSource = ESceneCaptureSource::SCS_SceneColorHDR;
	}	
	if (ForegroundCapture->CaptureSource == ESceneCaptureSource::SCS_SceneColorHDRNoAlpha ||
		ForegroundCapture->CaptureSource == ESceneCaptureSource::SCS_SceneColorSceneDepth)
	{
		ForegroundCapture->CaptureSource = ESceneCaptureSource::SCS_SceneColorHDR;
	}

	if (FullCapture->CaptureSource != ForegroundCapture->CaptureSource)
	{
		MIXCAST_ERROR("The Full and Foreground SceneCaptureComponents should be set to use the same CaptureSource type. Using the Full Capture's value.");
		ForegroundCapture->CaptureSource = FullCapture->CaptureSource;
	}

	switch (FullCapture->CaptureSource)
	{
	case ESceneCaptureSource::SCS_SceneColorHDR:
	case ESceneCaptureSource::SCS_FinalColorLDR:
#if UE_4_26_OR_LATER
	case ESceneCaptureSource::SCS_FinalColorHDR:
	case ESceneCaptureSource::SCS_FinalToneCurveHDR:
#endif
		break;

	default:
		MIXCAST_ERROR("MixCast's SceneCaptureComponents must be set to either a Scene or Final Color CaptureSource type. Using Final Color as fallback.");
		FullCapture->CaptureSource = ESceneCaptureSource::SCS_FinalColorLDR;
		ForegroundCapture->CaptureSource = ESceneCaptureSource::SCS_FinalColorLDR;
		break;
	}


#if !MIXCAST_SHADER_PASSES
	if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::PerPixel) {
		ForegroundCutoffMesh = NewObject<UForegroundCutoffMeshComponent>(this);
		ForegroundCutoffMesh->Rename(TEXT("Foreground Cutoff"));
		ForegroundCutoffMesh->AttachToCamera(this);
		ForegroundCutoffMesh->RegisterComponent();
	}
#endif

	framePipe = new FExpFrameSender(this);
	FString queueId = SharedTexIds::Cameras::RequestedExpFrames::Get(Identifier);
	SharedQueueConsumer<ExpFrame>** foundQueue = cachedFrameQueues.Find(queueId);
	if (foundQueue == nullptr)
	{
		std::wstring srcId = StringUtils::ToStdWString(queueId);
		requestedFrames = new SharedQueueConsumer<ExpFrame>(srcId.c_str());
		cachedFrameQueues.Add(queueId, requestedFrames);
	}
	else
		requestedFrames = *foundQueue;
}
UMixCastSceneCaptureComponent2D* AExpCamera::CreateCaptureComponent(bool isForeground)
{
	FName compName(isForeground ? "Foreground" : "Full");
	UMixCastSceneCaptureComponent2D* comp = CreateDefaultSubobject<UMixCastSceneCaptureComponent2D>(compName);
	comp->Initialize(isForeground);
	comp->SetupAttachment(CameraTransform, NAME_None);

	return comp;
}
void AExpCamera::CreateSliceCanvas(unsigned int width, unsigned int height, bool separateAlpha)
{
	LayerCanvas = NewObject<UMixCastSliceCanvas>(this);
	LayerCanvas->Initialize(width, height, separateAlpha);

	FullCapture->AssignLayersTarget(LayerCanvas->GetFullTex(), LayerCanvas->GetFullAlphaTex());
	ForegroundCapture->AssignLayersTarget(LayerCanvas->GetForegroundTex(), LayerCanvas->GetForegroundAlphaTex());
}
void AExpCamera::DestroySliceCanvas()
{
	if (FullCapture && FullCapture->IsValidLowLevel())
		FullCapture->AssignLayersTarget(NULL, NULL);
	if (ForegroundCapture && ForegroundCapture->IsValidLowLevel())
		ForegroundCapture->AssignLayersTarget(NULL, NULL);

	LayerCanvas = nullptr;
}
#if MIXCAST_SHADER_PASSES
void AExpCamera::CreateOpaqueLayerTarget(unsigned int width, unsigned int height)
{
	OpaqueTarget = NewObject<UTextureRenderTarget2D>(this, FName(TEXT("Opaque")));
	OpaqueTarget->bAutoGenerateMips = false;
	OpaqueTarget->TargetGamma = 2.2f;
	OpaqueTarget->InitCustomFormat(width, height, PF_FloatRGBA, false);
	OpaqueTarget->UpdateResourceImmediate();

	if (FullCapture->CaptureSource != SCS_SceneColorHDR)
	{
		ProcessedOpaqueColorTarget = NewObject<UTextureRenderTarget2D>(this, FName(TEXT("Processed Opaque Color")));
		ProcessedOpaqueColorTarget->bAutoGenerateMips = false;
		ProcessedOpaqueColorTarget->TargetGamma = 2.2f;
		ProcessedOpaqueColorTarget->InitCustomFormat(width, height, PF_FloatRGBA, false);
		ProcessedOpaqueColorTarget->UpdateResourceImmediate();

		OpaqueCanvas = NewObject<UMixCastOpaqueCanvas>(this);
		OpaqueCanvas->Initialize(OpaqueTarget, ProcessedOpaqueColorTarget);

		OpaqueProcessingCapture = NewObject<UOpaqueProcessingCaptureComponent>(this);
		OpaqueProcessingCapture->Rename(TEXT("Opaque Processing"));
		OpaqueProcessingCapture->Initialize();
		OpaqueProcessingCapture->OpaqueTarget = OpaqueTarget;
		OpaqueProcessingCapture->TextureTarget = ProcessedOpaqueColorTarget;
		OpaqueProcessingCapture->RegisterComponent();
	}

	FullCapture->AssignOpaqueTargets(OpaqueTarget, ProcessedOpaqueColorTarget);
}
void AExpCamera::DestroyOpaqueLayerTarget()
{
	if (FullCapture && FullCapture->IsValidLowLevel())
		FullCapture->AssignOpaqueTargets(NULL, NULL);

	OpaqueTarget = NULL;
	ProcessedOpaqueColorTarget = NULL;
	if (OpaqueProcessingCapture) 
	{
		OpaqueProcessingCapture->Rename(TEXT(""));
		OpaqueProcessingCapture->DestroyComponent();
		OpaqueProcessingCapture = NULL;
	}
	OpaqueCanvas = NULL;
}
#endif

void AExpCamera::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	if (framePipe) {
		delete framePipe;
		framePipe = nullptr;
	}

	if (FullCapture && FullCapture->IsValidLowLevel()) {
		FullCapture->OnCaptureBegin().Remove(onPreFullRender);
		FullCapture->OnCaptureEnd().Remove(onPostFullRender);
		FullCapture->AssignLayersTarget(NULL, NULL);
		FullCapture->DestroyComponent();
	}
	FullCapture = nullptr;

	if (ForegroundCapture && ForegroundCapture->IsValidLowLevel()) {
		ForegroundCapture->OnCaptureBegin().Remove(onPreForegroundRender);
		ForegroundCapture->OnCaptureEnd().Remove(onPostForegroundRender);
		ForegroundCapture->AssignLayersTarget(NULL, NULL);
		ForegroundCapture->DestroyComponent();
	}
	ForegroundCapture = nullptr;

	DestroySliceCanvas();

	Super::EndPlay(EndPlayReason);
}

void AExpCamera::CaptureIfNeeded()
{
	extern ENGINE_API float GAverageMS;

	unsigned long long curTime = TimeUtils::GetMixCastTimestamp();
	unsigned long long syncWindowEnd = curTime + (unsigned long)(0.5f * GAverageMS * 10000);

	bool haveFrameToRender = false;

	requestedFrames->EnqueueFenceUpdates();

	if (receivedLastFrameTooEarly)
	{
		if (lastReceivedFrame.syncTime <= syncWindowEnd)
		{
			receivedLastFrameTooEarly = false;
			latestFrameInfo = lastReceivedFrame;
			unsigned long long syncWindowStart = curTime - (unsigned long)(0.75f * TimeUtils::TicksPerSecond / latestFrameInfo.camFramerate);

			if (latestFrameInfo.syncTime >= syncWindowStart)
				haveFrameToRender = true;
			else
			{
				if (LOG_CAM_FRAME_DROPS) {
					MIXCAST_WARNING("Went from Too early to Too late to render frame %i, missed by %.2fms",
						latestFrameInfo.frameIndex,
						1000.0f * (float)(syncWindowStart - latestFrameInfo.syncTime) / TimeUtils::TicksPerSecond);
				}
				requestedFrames->MarkEmptied();
			}
		}
	}

	if (!receivedLastFrameTooEarly)
	{
		while (!haveFrameToRender && requestedFrames->WaitUntilNextFilled(0))
		{
			requestedFrames->Read(lastReceivedFrame);

			if (lastReceivedFrame.syncTime > syncWindowEnd)
			{
				receivedLastFrameTooEarly = true;
				break;
			}

			latestFrameInfo = lastReceivedFrame;
			unsigned long long syncWindowStart = curTime - (unsigned long)(0.75f * TimeUtils::TicksPerSecond / latestFrameInfo.camFramerate);

			if (latestFrameInfo.syncTime < syncWindowStart)
			{
				if (LOG_CAM_FRAME_DROPS) {
					MIXCAST_WARNING("Received Too late to render frame %i, missed by %.2fms",
						latestFrameInfo.frameIndex,
						1000.0f * (float)(syncWindowStart - latestFrameInfo.syncTime) / TimeUtils::TicksPerSecond);
				}
				requestedFrames->MarkEmptied();
				continue;
			}

			haveFrameToRender = true;
		}
	}

	if (!haveFrameToRender)
		return;

	CaptureFrame();
}
void AExpCamera::CaptureFrame()
{
	if (LayerCanvas != NULL && 
		(latestFrameInfo.camWidth != LayerCanvas->GetLayerWidth() || latestFrameInfo.camHeight != LayerCanvas->GetLayerHeight()))
		DestroySliceCanvas();
	if (LayerCanvas == NULL)
		CreateSliceCanvas(latestFrameInfo.camWidth, latestFrameInfo.camHeight, FullCapture->CaptureSource != ESceneCaptureSource::SCS_SceneColorHDR);

#if MIXCAST_SHADER_PASSES
	if (OpaqueTarget != NULL &&
		(latestFrameInfo.camWidth != OpaqueTarget->GetSurfaceWidth() || latestFrameInfo.camHeight != OpaqueTarget->GetSurfaceHeight() || !latestFrameInfo.HasCamFlag(ExpCamFlagBit::SeparateOpaque)))
		DestroyOpaqueLayerTarget();
	if (OpaqueTarget == NULL && latestFrameInfo.HasCamFlag(ExpCamFlagBit::SeparateOpaque))
		CreateOpaqueLayerTarget(latestFrameInfo.camWidth, latestFrameInfo.camHeight);
#endif

	UpdateTransform(latestFrameInfo.camPos, latestFrameInfo.camRot);
	UpdateCameraParameters();
	ResetHiddenActors();

	TriggerSceneCaptures();
}
void AExpCamera::UpdateTransform(FVector camPos, FQuat camRot)
{
	float WorldToMeters = GetWorld()->GetWorldSettings()->WorldToMeters;
	RootComponent->SetRelativeLocationAndRotation(MathUtil::MixCastPositionToUnreal(camPos, WorldToMeters), MathUtil::MixCastRotationToUnreal(camRot));
	RootComponent->UpdateComponentToWorld();
}
void AExpCamera::UpdateCameraParameters()
{
	//update field of view
	float hFoV = MathUtil::VerticalFoVToHorizontal(latestFrameInfo.camFoV, (float)latestFrameInfo.camWidth / latestFrameInfo.camHeight);
	if (hFoV != FullCapture->FOVAngle) {
		FullCapture->FOVAngle = hFoV;
		ForegroundCapture->FOVAngle = hFoV;
	}

#if !MIXCAST_SHADER_PASSES
	//Update foreground clipping plane
	if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::Planar) 
	{
		FTransform camTrans = RootComponent->GetComponentTransform();
		FVector camForward = camTrans.TransformVector(FVector::ForwardVector);
		FVector clipPos = camTrans.TransformPosition(FVector::ForwardVector * latestFrameInfo.occlusionApproxDepth * GetWorld()->GetWorldSettings()->WorldToMeters);

		ForegroundCapture->ClipPlaneBase = clipPos;
		ForegroundCapture->ClipPlaneNormal = -camForward;
	}
#endif

	//Update how background is rendered
	if (latestFrameInfo.renderFull)
	{
		bool fullIsTranslucent = latestFrameInfo.HasCamFlag(ExpCamFlagBit::Translucent);
		if (fullRenderSetAsTransparent != fullIsTranslucent)
		{
			if (fullIsTranslucent) {
				FullCapture->ShowFlags.SetFog(false);
#if !UE_4_24_OR_LATER
				FullCapture->ShowFlags.SetAtmosphericFog(false);
#endif
			}
			fullRenderSetAsTransparent = fullIsTranslucent;
		}
	}

	//Update post-processing settings
	APlayerController* vrPlayer = GEngine->GetFirstLocalPlayerController(GetWorld());
	AActor* camActor = vrPlayer && vrPlayer->PlayerCameraManager ?
		vrPlayer->PlayerCameraManager->ViewTarget.Target :
		nullptr;
	UCameraComponent* camComp = camActor ?
		Cast<UCameraComponent>(camActor->GetComponentByClass(UCameraComponent::StaticClass())) :
		nullptr;

	if (camComp)
	{
		FullCapture->PostProcessSettings = camComp->PostProcessSettings;
		FullCapture->PostProcessBlendWeight = camComp->PostProcessBlendWeight;

		ForegroundCapture->PostProcessSettings = camComp->PostProcessSettings;
		ForegroundCapture->PostProcessBlendWeight = camComp->PostProcessBlendWeight;
	}
}
void AExpCamera::ResetHiddenActors()
{
	FullCapture->HiddenActors.Empty();
	ForegroundCapture->HiddenActors.Empty();

	for (auto& cutoffMesh : UForegroundCutoffMeshComponent::Active)
	{
		AActor* owner = cutoffMesh->GetOwner();
		FullCapture->HiddenActors.Add(owner);
		if (cutoffMesh != ForegroundCutoffMesh)
		{
			ForegroundCapture->HiddenActors.Add(owner);
		}
	}
}
void AExpCamera::TriggerSceneCaptures()
{
	check(rendersLeft == 0);
	if (latestFrameInfo.renderFull)
		rendersLeft++;
	if (latestFrameInfo.renderForeground)
		rendersLeft++;

	FrameStarted.Broadcast(this);

	if (latestFrameInfo.renderFull)
		FullCapture->CaptureFrameDeferred(latestFrameInfo);
	if (latestFrameInfo.renderForeground)
		ForegroundCapture->CaptureFrameDeferred(latestFrameInfo);
}

void AExpCamera::HandleSceneCaptureStarted(UMixCastSceneCaptureComponent2D* sceneCapture)
{
}
void AExpCamera::HandleSceneCaptureEnded(UMixCastSceneCaptureComponent2D* sceneCapture)
{
	if (rendersLeft <= 0) {
		MIXCAST_WARNING("Didn't ask for render!");
		return;
	}
	if (--rendersLeft == 0) {
		AtlasSlices();

		if (OpaqueCanvas) {
			OpaqueProcessingCapture->CaptureSource = FullCapture->CaptureSource;
			OpaqueProcessingCapture->PostProcessSettings = FullCapture->PostProcessSettings;
			OpaqueProcessingCapture->PostProcessBlendWeight = FullCapture->PostProcessBlendWeight;
			OpaqueProcessingCapture->CaptureScene();

			OpaqueCanvas->FastUpdateResource();
		}

		requestedFrames->EnqueueFenceInsert();

		FrameEnded.Broadcast(this);
	}
}

void AExpCamera::AtlasSlices()
{
	if (!LayerCanvas)
		return;

	LayerCanvas->blitFull = latestFrameInfo.renderFull;
	LayerCanvas->blitForeground = latestFrameInfo.renderForeground;
	LayerCanvas->renderTranslucent = latestFrameInfo.HasCamFlag(ExpCamFlagBit::Translucent) || latestFrameInfo.HasCamFlag(ExpCamFlagBit::SeparateOpaque);
	LayerCanvas->FastUpdateResource();
}

void AExpCamera::SetActorHidden(AActor* actor)
{
	FullCapture->HiddenActors.Add(actor);
	ForegroundCapture->HiddenActors.Add(actor);
}

float AExpCamera::GetNearClipPlane()
{
#if UE_4_23_OR_LATER
	if (FullCapture->bOverride_CustomNearClippingPlane)
		return FullCapture->CustomNearClippingPlane;
	else
#endif
		return GNearClippingPlane;
}
