/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "ExpViewfinder.h"

#include "MixCast.h"
#include "Data/SharedTexIds.h"

#include "Engine/Engine.h"
#include "UObject/ConstructorHelpers.h"
#include "GameFramework/WorldSettings.h"

#include <string>

AExpViewfinder::AExpViewfinder(const FObjectInitializer &ObjectInitializer)
{
	static ConstructorHelpers::FObjectFinder<UStaticMesh> Find_PlaneMeshAsset(TEXT("StaticMesh'/Engine/BasicShapes/Plane'"));
	if (Find_PlaneMeshAsset.Succeeded())
	{
		PlaneMesh = Find_PlaneMeshAsset.Object;
	}

	static ConstructorHelpers::FObjectFinder<UMaterial> Find_DispayViewfinderMat(TEXT("Material'/MixCast/DisplayViewfinder.DisplayViewfinder'"));
	if (Find_DispayViewfinderMat.Succeeded())
	{
		DisplayMat = UMaterialInstanceDynamic::Create(
			Cast<UMaterialInterface>(Find_DispayViewfinderMat.Object),
			this,
			TEXT("DisplayViewfinder")
		);
	}

	SetActorEnableCollision(false);
	SpawnCollisionHandlingMethod = ESpawnActorCollisionHandlingMethod::AlwaysSpawn;

	ViewfinderTransform = CreateDefaultSubobject<USceneComponent>(FName("Transform"));
	SetRootComponent(ViewfinderTransform);

	DisplayMesh = CreateDefaultSubobject<UStaticMeshComponent>(FName("Display Mesh"));
	DisplayMesh->SetStaticMesh(PlaneMesh);
	DisplayMesh->SetMaterial(0, DisplayMat);
	DisplayMesh->SetupAttachment(ViewfinderTransform);
	DisplayMesh->SetRelativeLocation(FVector::ZeroVector);
	DisplayMesh->SetRelativeRotation(FQuat::MakeFromEuler(FVector(-90, 0, 90)));
	DisplayMesh->SetCastShadow(false);

	PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.TickGroup = TG_PostUpdateWork;
	PrimaryActorTick.bStartWithTickEnabled = true;
}

void AExpViewfinder::BeginPlay()
{
	Super::BeginPlay();

	DisplayReceiver = new SharedTextureReceiver(SharedTexIds::Viewfinders::Display::Get(Identifier));
	DisplayReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpViewfinder::HandleTextureChanged);
	HandleTextureChanged();
}
void AExpViewfinder::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	if (DisplayMesh) {
		if (DisplayMesh->IsValidLowLevel()) {
			DisplayMesh->DestroyComponent();
		}
		DisplayMesh = nullptr;
	}

	if (DisplayReceiver)
	{
		DisplayReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpViewfinder::HandleTextureChanged);
		delete DisplayReceiver;
		DisplayReceiver = nullptr;
	}

	Super::EndPlay(EndPlayReason);
}

void AExpViewfinder::HandleTextureChanged()
{
	UTexture* tex = DisplayReceiver->GetTexture();
	if (tex)
		tex->SRGB = 1;
	DisplayMat->SetTextureParameterValue(FName("DisplayTex"), tex);
}

void AExpViewfinder::Tick(float DeltaSeconds)
{
	float worldToMeters = GetWorld()->GetWorldSettings()->WorldToMeters;

	Viewfinder* viewfinderData = SdkData::GetViewfinderById(Identifier);
	if (viewfinderData)
	{
		FVector camPos = ThriftUtil::ThriftPositionToUnreal(viewfinderData->currentPosition, worldToMeters);
		FQuat camRot = ThriftUtil::ThriftQuatToUnreal(viewfinderData->currentRotation);
		RootComponent->SetRelativeLocationAndRotation(camPos, camRot);
		RootComponent->UpdateComponentToWorld();
	}

	DisplayReceiver->EnqueueUpdate();

	if (DisplayReceiver->GetTexture()) {
		FVector scale = ThriftUtil::ThriftScaleToUnreal(viewfinderData->currentScale);
		UTexture* tex = DisplayReceiver->GetTexture();
		scale.X *= tex->GetSurfaceWidth() / tex->GetSurfaceHeight();
		DisplayMesh->SetRelativeScale3D(scale);
	}
}