/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "MixCast_Oculus.h"

#include "MixCastDefines.h"
#include "MixCast.h"
#include "MixCastSdkActor.h"
#include "MixCast/Private/Tracking/TrackingUtil.h"
#include "MixCast_OculusLog.h"

#include "CoreMinimal.h"
#include "GameFramework/WorldSettings.h"
#include "HeadMountedDisplayFunctionLibrary.h"

#if UE_4_25_OR_LATER
#include "OculusHMD/Private/OculusPluginWrapper.h"
#include "OculusHMD/Public/OculusFunctionLibrary.h"
#include "OculusHMD/Private/OculusHMDPrivate.h"
#include "OculusHMD/Private/OculusHMDModule.h"
#elif UE_4_20_OR_LATER
#include "Oculus/OVRPlugin/OVRPlugin/Include/OVR_Plugin.h"
#include "OculusHMD/Public/OculusFunctionLibrary.h"
#include "OculusHMD/Private/OculusHMDPrivate.h"
#include "OculusHMD/Private/OculusHMDModule.h"
#else
#include "OVRPlugin/Include/OVR_Plugin.h"
#include "Public/OculusFunctionLibrary.h"
#include "Private/OculusHMDPrivate.h"
#include "Private/OculusHMDModule.h"
#endif

IMPLEMENT_MODULE(FMixCast_Oculus, MixCast_Oculus)

void FMixCast_Oculus::StartupModule()
{
	IMixCast& mod = IMixCast::Get();
	activeStateDelegate = mod.OnActiveChanged().AddRaw(this, &FMixCast_Oculus::HandleMixCastActiveChanged);
	trackingSpaceChangedDelegate = mod.OnTrackingSpaceChanged().AddRaw(this, &FMixCast_Oculus::HandleTrackingSpaceChanged);
	beforeFrameDelegate = mod.OnBeforeFrame().AddRaw(this, &FMixCast_Oculus::HandleMixCastAboutToRender);
}
void FMixCast_Oculus::ShutdownModule()
{
	if (IMixCast::IsAvailable()) {
		IMixCast& mod = IMixCast::Get();
		mod.OnActiveChanged().Remove(activeStateDelegate);
		mod.OnTrackingSpaceChanged().Remove(trackingSpaceChangedDelegate);
		mod.OnBeforeFrame().Remove(beforeFrameDelegate);
	}
}
void FMixCast_Oculus::HandleMixCastActiveChanged()
{
	if (IMixCast::Get().IsActive())
		HandleTrackingSpaceChanged();
}
void FMixCast_Oculus::HandleTrackingSpaceChanged()
{
	FTransform newTrans;
	if (CalculateTrackingToPlayAreaTransform(newTrans))
		trackingToPlayArea = newTrans;
}

void FMixCast_Oculus::HandleMixCastAboutToRender()
{
	if (TrackingUtil::GetCurrentTrackingType(true) != TrackingSource::OCULUS)
		return;

	FTransform hmdToWorld, baseToHmd, trackingToBase;
	bool success = CalculateHmdToWorldTransform(hmdToWorld);
	if (!success)
		return;

	success = CalculateBaseToHmdTransform(baseToHmd);
	if (!success)
		return;

	success = CalculateTrackingToBaseTransform(trackingToBase);
	if (!success)
		return;

	if (DetectBoundaryChanged())
		HandleTrackingSpaceChanged();

	FTransform playAreaToWorld = trackingToPlayArea * trackingToBase.Inverse() * baseToHmd.Inverse() * hmdToWorld;

	AMixCastSdkActor::SetWorldTransform(playAreaToWorld);
}

bool FMixCast_Oculus::DetectBoundaryChanged()
{
	static ovrpVector3f playRect[4];
	int NumPoints = 0;
#if UE_4_25_OR_LATER
	if (!OVRP_SUCCESS(FOculusHMDModule::GetPluginWrapper().GetBoundaryGeometry3(ovrpBoundary_PlayArea, playRect, &NumPoints)))
		return false;
#else
	if (!OVRP_SUCCESS(ovrp_GetBoundaryGeometry3(ovrpBoundary_PlayArea, playRect, &NumPoints)))
		return false;
#endif

	FVector newPos = FMath::Lerp(OculusHMD::ToFVector(playRect[0]), OculusHMD::ToFVector(playRect[2]), 0.5f);
	FQuat newRot = (OculusHMD::ToFVector(playRect[1]) - OculusHMD::ToFVector(playRect[2])).ToOrientationQuat();
	if (FVector::DistSquared(newPos, cachedPlayRectPos) > 0.01f * 0.01f || FMath::RadiansToDegrees(newRot.AngularDistance(cachedPlayRectRot)) > 0.5f) {
		cachedPlayRectPos = newPos;
		cachedPlayRectRot = newRot;
		return true;
	}

	return false;
}

bool FMixCast_Oculus::CalculateHmdToWorldTransform(FTransform& trans)
{
#if UE_4_19_OR_LATER
	if (!GEngine->XRSystem.IsValid())
		return false;

	trans = GEngine->XRSystem->GetTrackingToWorldTransform();
	return true;
#else
	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	if (curWorld == NULL)
		return false;

	APlayerController* localPlayerController = GEngine->GetFirstLocalPlayerController(curWorld);
	if (localPlayerController == NULL)
		return false;

	APlayerCameraManager* playerCamManager = localPlayerController->PlayerCameraManager;
	if (playerCamManager == NULL)
		return false;

	trans = FTransform(playerCamManager->GetCameraRotation(), playerCamManager->GetCameraLocation());
	return true;
#endif
}

bool FMixCast_Oculus::CalculateBaseToHmdTransform(FTransform& trans)
{
#if UE_4_19_OR_LATER
	trans = FTransform::Identity;	//can get the tracking origin on its own now
	return true;
#elif UE_4_18_OR_LATER
	FQuat trackedHmdRot;
	FVector trackedHmdPos;
	if (!GEngine->XRSystem.IsValid() || !GEngine->XRSystem->GetCurrentPose(IXRTrackingSystem::HMDDeviceId, trackedHmdRot, trackedHmdPos))
		return false;

	trans = FTransform(trackedHmdRot, trackedHmdPos);
	return true;
#else
	if (!GEngine->HMDDevice.IsValid() || !GEngine->HMDDevice->HasValidTrackingPosition())
		return false;

	FQuat trackedHmdRot;
	FVector trackedHmdPos;
	GEngine->HMDDevice->GetCurrentOrientationAndPosition(trackedHmdRot, trackedHmdPos);
	trans = FTransform(trackedHmdRot, trackedHmdPos);
	return true;
#endif
}

bool FMixCast_Oculus::CalculateTrackingToBaseTransform(FTransform& trans)
{
	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	float worldScale = curWorld != nullptr ? 
		curWorld->GetWorldSettings()->WorldToMeters :
		100;

	FVector basePos;
	FRotator baseRot;
	UOculusFunctionLibrary::GetBaseRotationAndBaseOffsetInMeters(baseRot, basePos);

	trans = FTransform(baseRot, basePos * worldScale);
	return true;
}

bool FMixCast_Oculus::CalculateTrackingToPlayAreaTransform(FTransform& trans)
{
	if (!FOculusHMDModule::IsAvailable())
		return false;

	FOculusHMDModule& ocMod = FOculusHMDModule::Get();
	if (!ocMod.IsOVRPluginAvailable() || !ocMod.IsHMDConnected())
		return false;

#if UE_4_25_OR_LATER
	if (FOculusHMDModule::GetPluginWrapper().GetInitialized() == ovrpBool_False)
		return false;
#else
	if (!ovrp_GetInitialized())
		return false;
#endif

	int NumPoints = 0;
#if UE_4_25_OR_LATER
	if (!OVRP_SUCCESS(FOculusHMDModule::GetPluginWrapper().GetBoundaryGeometry3(ovrpBoundary_Outer, NULL, &NumPoints)))
		return false;
#else
	if (!OVRP_SUCCESS(ovrp_GetBoundaryGeometry3(ovrpBoundary_Outer, NULL, &NumPoints)))
		return false;
#endif

	int BufferSize = NumPoints;
	ovrpVector3f* BoundaryPoints = new ovrpVector3f[BufferSize];

#if UE_4_25_OR_LATER
	if (!OVRP_SUCCESS(FOculusHMDModule::GetPluginWrapper().GetBoundaryGeometry3(ovrpBoundary_Outer, BoundaryPoints, &NumPoints)))
	{
		delete[] BoundaryPoints;
		return false;
	}
#else
	if (!OVRP_SUCCESS(ovrp_GetBoundaryGeometry3(ovrpBoundary_Outer, BoundaryPoints, &NumPoints)))
	{
		delete[] BoundaryPoints;
		return false;
	}
#endif
	NumPoints = FMath::Min(BufferSize, NumPoints);

	UWorld* curWorld = ((FMixCast*)&IMixCast::Get())->GetWorld();
	float worldScale = curWorld != nullptr ? 
		curWorld->GetWorldSettings()->WorldToMeters :
		100;

	FVector averagePos(0);
	for (int i = 0; i < NumPoints; i++)
		averagePos += OculusHMD::ToFVector(BoundaryPoints[i]);
	averagePos /= NumPoints;
	averagePos *= worldScale;

	FVector forwardPos = OculusHMD::ToFVector(BoundaryPoints[0]) * worldScale;

	delete[] BoundaryPoints;

	FQuat trackingCenterRot = (forwardPos - averagePos).ToOrientationQuat();
	trans = FTransform(trackingCenterRot, averagePos);
	return true;
}