/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "WriteOcclusion.h"
#if MIXCAST_SHADER_PASSES
#include "CommonRenderResources.h"
#include "Containers/DynamicRHIResourceArray.h"
#include "Runtime/Renderer/Private/PostProcess/SceneRenderTargets.h"

class FWriteOcclusionVertexBuffer : public FVertexBuffer
{
public:
	/** Initialize the RHI for this rendering resource */
	void InitRHI()
	{
		TResourceArray<FFilterVertex, VERTEXBUFFER_ALIGNMENT> Vertices;
		Vertices.SetNumUninitialized(6);

		Vertices[0].Position = FVector4(-1, 1, 0, 1);
		Vertices[0].UV = FVector2D(0, 0);

		Vertices[1].Position = FVector4(1, 1, 0, 1);
		Vertices[1].UV = FVector2D(1, 0);

		Vertices[2].Position = FVector4(-1, -1, 0, 1);
		Vertices[2].UV = FVector2D(0, 1);

		Vertices[3].Position = FVector4(1, -1, 0, 1);
		Vertices[3].UV = FVector2D(1, 1);

		// Create vertex buffer. Fill buffer with initial data upon creation
		FRHIResourceCreateInfo CreateInfo(&Vertices);
		VertexBufferRHI = RHICreateVertexBuffer(Vertices.GetResourceDataSize(), BUF_Static, CreateInfo);
	}
};
TGlobalResource<FWriteOcclusionVertexBuffer> GOcclusionVertexBuffer;

class FWriteOcclusionVS : public FGlobalShader
{
	DECLARE_GLOBAL_SHADER(FWriteOcclusionVS);

public:
	FWriteOcclusionVS() { }
	FWriteOcclusionVS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{
		BufferRatioParam.Bind(Initializer.ParameterMap, TEXT("BufferRatio"));
	}

#if !UE_4_25_OR_LATER
	virtual bool Serialize(FArchive & Ar) override
	{
		bool bShaderHasOutdatedParameters = FGlobalShader::Serialize(Ar);
		Ar << BufferRatioParam;
		return bShaderHasOutdatedParameters;
	}
#endif

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters & Parameters)
	{
		return true;
	}

protected:
#if UE_4_25_OR_LATER
	LAYOUT_FIELD(FShaderParameter, BufferRatioParam);
#else
	FShaderParameter BufferRatioParam;
#endif

public:

	void SetParameters(FRHICommandList & RHICmdList, TShaderMapRef<FWriteOcclusionVS> ShaderRef, FSceneView & InView, FVector4 BufferRatio)
	{
#if !UE_4_25_OR_LATER
		FRHIVertexShader* ShaderRHI = GetVertexShader();
#else
		FRHIVertexShader* ShaderRHI = ShaderRef.GetVertexShader();
#endif

		FGlobalShader::SetParameters<FViewUniformShaderParameters>(RHICmdList, ShaderRHI, InView.ViewUniformBuffer);
		SetShaderValue(RHICmdList, ShaderRHI, BufferRatioParam, BufferRatio);
	}
};

template<int32 NumOutputs>
class FWriteOcclusionPS : public FGlobalShader
{
public:
	DECLARE_GLOBAL_SHADER(FWriteOcclusionPS);

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
	{
		return IsFeatureLevelSupported(Parameters.Platform, ERHIFeatureLevel::SM5);
	}

	static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
	{
		FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
		OutEnvironment.SetDefine(TEXT("NUM_OUTPUTS"), NumOutputs);
	}

protected:
#if UE_4_25_OR_LATER
	LAYOUT_FIELD(FShaderResourceParameter, DepthTexture);
	LAYOUT_FIELD(FShaderResourceParameter, DepthTextureSampler);
	LAYOUT_FIELD(FShaderParameter, DepthMultiplier);
#else
	FShaderResourceParameter DepthTexture;
	FShaderResourceParameter DepthTextureSampler;
	FShaderParameter DepthMultiplier;
#endif

public:
	FWriteOcclusionPS() { }
	FWriteOcclusionPS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{
		DepthTexture.Bind(Initializer.ParameterMap, TEXT("DepthTexture"));
		DepthTextureSampler.Bind(Initializer.ParameterMap, TEXT("DepthTextureSampler"));
		DepthMultiplier.Bind(Initializer.ParameterMap, TEXT("DepthMultiplier"));
	}

#if !UE_4_25_OR_LATER
	virtual bool Serialize(FArchive& Ar) override
	{
		bool bShaderHasOutdatedParameters = FGlobalShader::Serialize(Ar);
		Ar << DepthTexture << DepthTextureSampler << DepthMultiplier;
		return bShaderHasOutdatedParameters;
	}
#endif

	void SetParameters(FRHICommandList& RHICmdList, TShaderMapRef<FWriteOcclusionPS<NumOutputs>> ShaderRef, FSceneView& InView, FRHITexture* DepthTextureRHI, float DepthMultiplierValue)
	{
		FRHISamplerState* SamplerStateRHI = TStaticSamplerState<SF_Bilinear>::GetRHI();
#if !UE_4_25_OR_LATER
		FRHIPixelShader* ShaderRHI = GetPixelShader();
#else
		FRHIPixelShader* ShaderRHI = ShaderRef.GetPixelShader();
#endif

		FGlobalShader::SetParameters<FViewUniformShaderParameters>(RHICmdList, ShaderRHI, InView.ViewUniformBuffer);
		SetTextureParameter(RHICmdList, ShaderRHI, DepthTexture, DepthTextureSampler, SamplerStateRHI, DepthTextureRHI);
		SetShaderValue(RHICmdList, ShaderRHI, DepthMultiplier, DepthMultiplierValue);
	}
};

IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionVS, "/Plugin/MixCast/WriteOcclusion.usf", "MainVS", SF_Vertex);

IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<1>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<2>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<3>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<4>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<5>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<6>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<7>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FWriteOcclusionPS<8>, "/Plugin/MixCast/WriteOcclusion.usf", "MainPS", SF_Pixel);


void FMixCastShader_WriteOcclusion::Execute(FRHICommandListImmediate& RHICmdList, FSceneView& InView, bool basePass, FRHITexture* DepthTextureRHI, float DepthMultiplierValue)
{
	FSceneRenderTargets& SceneContext = FSceneRenderTargets::Get(RHICmdList);

	if (RHICmdList.IsInsideRenderPass())
		RHICmdList.EndRenderPass();

	//switch to new render pass
#if UE_4_25_OR_LATER
	FRHIRenderPassInfo RPInfo;
	int32 velIndex = -1;
#if UE_4_26_OR_LATER
	SceneContext.FillGBufferRenderPassInfo(ERenderTargetLoadAction::ELoad, RPInfo, velIndex);
#else
	int32 tangentIndex = -1;
	SceneContext.FillGBufferRenderPassInfo(ERenderTargetLoadAction::ELoad, RPInfo, velIndex, tangentIndex);
#endif
	RPInfo.DepthStencilRenderTarget.DepthStencilTarget = SceneContext.GetSceneDepthSurface();
	RPInfo.DepthStencilRenderTarget.ExclusiveDepthStencil = FExclusiveDepthStencil::DepthWrite_StencilNop;
	RPInfo.DepthStencilRenderTarget.Action = MakeDepthStencilTargetActions(ERenderTargetActions::Load_Store, ERenderTargetActions::DontLoad_DontStore);
	RHICmdList.BeginRenderPass(RPInfo, TEXT("MixCast-WriteOcclusion"));
#else
	SceneContext.BeginRenderingGBuffer(RHICmdList, ERenderTargetLoadAction::ELoad, ERenderTargetLoadAction::ELoad, FExclusiveDepthStencil::DepthWrite_StencilNop, false);
#endif


	auto ShaderMap = GetGlobalShaderMap(InView.GetFeatureLevel());

	TShaderMapRef<FWriteOcclusionVS> VertexShader(ShaderMap);

	//prepare state
	FGraphicsPipelineStateInitializer GraphicsPSOInit;
	RHICmdList.ApplyCachedRenderTargets(GraphicsPSOInit);
	GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();
	GraphicsPSOInit.RasterizerState = TStaticRasterizerState<>::GetRHI();
	GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<true, CF_DepthNearOrEqual>::GetRHI();

	GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = GFilterVertexDeclaration.VertexDeclarationRHI;
	GraphicsPSOInit.PrimitiveType = PT_TriangleStrip;
#if UE_4_25_OR_LATER
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = VertexShader.GetVertexShader();
#else
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = GETSAFERHISHADER_VERTEX(*VertexShader);
#endif

	//set pixel shader
	uint32 NumActiveRenderTargets = SceneContext.GetNumGBufferTargets();
	switch (NumActiveRenderTargets)
	{
	case 5:
	{
		TShaderMapRef<FWriteOcclusionPS<5>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	case 6:
	{
		TShaderMapRef<FWriteOcclusionPS<6>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	default:
	case 1:
	{
		TShaderMapRef<FWriteOcclusionPS<1>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	}

	FPooledRenderTargetDesc Desc = SceneContext.GetSceneColor()->GetDesc();
	FVector4 bufferRatio((float)DepthTextureRHI->GetTexture2D()->GetSizeX() / Desc.Extent.X,
		(float)DepthTextureRHI->GetTexture2D()->GetSizeY() / Desc.Extent.Y);

	VertexShader->SetParameters(RHICmdList, VertexShader, InView,
		bufferRatio);

	//dispatch draw
	RHICmdList.SetViewport(0, 0, 0, Desc.Extent.X, Desc.Extent.Y, 1);
	RHICmdList.SetStreamSource(0, GOcclusionVertexBuffer.VertexBufferRHI, 0);
	RHICmdList.DrawPrimitive(0, 2, 1);

#if UE_4_26_OR_LATER
	if (!basePass)
		RHICmdList.EndRenderPass();
#endif
}
#endif